{****************************************************************************}
{*                                                                          *}
{*  SSSS            t               EEEEE                                   *}
{* S     y  y  sss ttt  ee  mmm mm  E     x  x                              *}
{*  SSS  y  y s__   t  e  e m  m  m EEEE   \/                               *}
{*     S y  y    s  t  e e  m  m  m E      /\                               *}
{* SSSS   yyy sss    t  eee m  m  m EEEEE x  x UNIT                         *}
{*          y                                                               *}
{*       yyy                                                                *}
{****************************************************************************}
{* This unit is designed to greatly enhance the TP7/BP7 capabilities by     *}
{* adding preemptive multitasking, nested exception handling, messaging,    *}
{* pipes, mutexes and a few other system enhancements like a 6ms precise    *}
{* timer or a FIFO queue for pointers.                                      *}
{*                                                                          *}
{* The most important CRT functions can also be found in this unit so that  *}
{* CRT doesn't have to be imported. The "self" function returns a pointer   *}
{* to the running thread; this is useful for message handling blocking      *}
{* operations on threads.                                                   *}
{*                                                                          *}
{* IMPORTANT NOTE: Due to DMPI host limitation the multithreading is only   *}
{*                 cooperative when using DPMI mode. Use Switch often.      *}
{****************************************************************************}
{* THREADS, MESSAGES AND PIPES                                              *}
{*                                                                          *}
{* Thread Setup: Derive an object from TThread and write your own (virtual) *}
{*               Execute method. Add as many dynamic functions as you like  *}
{*               with the following syntax:                                 *}
{*   procedure MyMessage(var Message: TMessage); virtual MessageID;         *}
{*                                                                          *}
{* Threads will automatically terminate and free themselves when reaching   *}
{* the end of the Execute procedure. Note that you can also "shoot" a       *}
{* thread by calling his Free method out of himself or another thread.      *}
{*                                                                          *}
{* Pipes will open two virtual text devices I and O for input and output.   *}
{* All the data written to a pipe will be held in a single container, but   *}
{* threads cannot receive the data they sent themselves. That allows you to *}
{* let multiple threads read and write to the same pipe. This can be useful *}
{* to find idle threads: connect all "client" threads to a pipe with one    *}
{* "server" thread. When a client is busy, he will not read the pipe. As    *}
{* soon as the server sends a command into the pipe, the first idle client  *}
{* will receive the command.                                                *}
{*                                                                          *}
{* Use messages to send specific data to another thread or object derived   *}
{* from MessageObject. Because TMessage is an object you can virtually send *}
{* any inforation or data. However, you must make sure that the destructor  *}
{* Done will free all the message data in order to avoid memory leaks. The  *}
{* destructor is automatically called after the message has been processed. *}
{****************************************************************************}
{* EXCEPTION HANDLING                                                       *}
{*                                                                          *}
{* The runtime errors are very bad for multithreading applications because  *}
{* they usually stop the whole program. If you use the exception handling   *}
{* provided in this unit you'll be able to write elegant, resource saving   *}
{* exception blocks. Each thread has its own exception root, and exception  *}
{* blocks can be nested. Any runtime error occuring in the exception blocks *}
{* will not stop the application but jump to the exception handler code.    *}
{* Exceptions can be re-raised in order to go one level further down.       *}
{*                                                                          *}
{* The syntax is easy. For example, for file access, you could write this:  *}
{*   Assign(F,'MYFILE');                                                    *}
{*   if Try then begin                                                      *}
{*     Reset(F);                                                            *}
{*     EndTry;                                                              *}
{*   end else begin                                                         *}
{*     if Exception=2 then                                                  *}
{*       Rewrite(F)                                                         *}
{*     else                                                                 *}
{*       Raise;                                                             *}
{*   end;                                                                   *}
{*   ...Use file here...                                                    *}
{*   Close(F);                                                              *}
{* Now what happens? If the file doesn't exist, an exception raises and the *}
{* ELSE block will be processed. If the error is "File not found", try to   *}
{* it; if an errors occures here or if it was another error we just jump to *}
{* the previous exception handler.                                          *}
{*                                                                          *}
{* Note the EndTry command before the "end else". This commanbd is needed   *}
{* to close the exception block; if you forget it you'll experience VERY    *}
{* strange results or crashes!                                              *}
{****************************************************************************}
{* DISCLAIMER: The author does not take any responsability for any damage   *}
{*             done by the use of this unit. It is provided "AS-IS" without *}
{*             any warranty of any kind.                                    *}
{****************************************************************************}
{* Author:  (C) by Arsne von Wyss                                          *}
{* Created: 17.02.1998                                                      *}
{* Version: 0.5 (TMutex untested)                                           *}
{****************************************************************************}
(* REFERENCE

TPointerFIFO:
		constructor Init;
		function Count: Word;      Pointers waiting in FIFO
		procedure Put(P: Pointer); Add Pointer to FIFO
		function Get: Pointer;     Remove pointer from FIFO
		destructor Done; virtual;  Destory FIFO

TMessage:
		Sender: PMessageObject;    Sender object
		Message: Word;             Message ID, needed for dynamic procedure call
		constructor Init;          Initialize Message
		destructor Done; virtual;  Destory Message

TMessageObject:
		constructor Init;          Initialize message object
		procedure PostMessage(AMessage: PMessage); Queue a previously created message
		procedure Perform(AMessage: PMessage); virtual; Process a specific message in the running thread.
		procedure ProcessMessages; Process queued messages. Usually used in a TThread.Execute loop.
		destructor Done; virtual;  Destroy message object

TThread:
		Name: string[31];          Name field; use as you wish
		constructor Init(const AName: string; AStackSize: Word; User: Pointer); Create and start thread
		procedure Execute(User: Pointer); virtual; Abstract execution routine
		procedure Stop;            Stop thread. Can be used as remote command
		procedure Start;           Restart thread. Can be used as remote command
		function Running: Boolean; Thread status (e.g. Stopped/Started)
		destructor Done; virtual;  Destroy thread

TPipe:
		I,O: Text;                 Virtual I/O text files; do not close or reassign!
		constructor Init;          Create new pipe
		function Empty: Boolean;   True if no data is pending for the current thread.
		destructor Done; virtual;  Destroy pipe

TMutex:
		constructor Init;          Create new mutex
		function Locked: Boolean;  Returns tur if the mutex is already in use
		procedure Lock;            Lock mutex; thread is suspended if needed until the Mutex gets free
		procedure UnLock;          Unlock mutex. Very important; else other threads will hang
		destructor Done; virtual;  Destroy Mutex; waiting threads are NOT re-enabled

const
	SwitchOk: Integer=1;         Forbid/Permit counter. You should only read it.

var
	Timer: Longint;              Timer in ms: read only; updates approx. every 6 ms
	Threads: Integer;            Counter of running threads; read only
	Main: TThread;               Virtual thread of main application
	Exception: Integer;          Exception code; equal to ExitCode

procedure Sound(Hz: Word);     CRT Sound
procedure NoSound;             CRT NoSound
procedure Delay(Ms: Longint);  CRT Delay, but gives away timeslices
function ReadKey: Char;        CRT ReadKey
function KeyPressed: Boolean;  CRT KeyPressed
procedure ClearKeys;           Clear keyboard buffer
function Try: Boolean;         Start exception block; use only in IF statments!
procedure EndTry;              End of exception block
procedure Raise;               Equal to RunError(Exception)
procedure Abort;               Equal to RunError(EAbort)
procedure Switch;              Give away current timeslice, allowing cooperative threading
procedure SwitchTo(AThread: PThread); Skip to AThread
procedure Forbid;              Stop task switching
procedure Permit;              Start task swithing
function self: PThread;        Currently running thread
*)

{$A+,B-,D+,E+,F-,G+,I-,L+,N+,P-,Q-,R-,S-,T-,V-,X+,Y+}

unit SystemEx;

interface

uses Objects;

type
{Pointer FIFO queue}
	PPointerFIFO=^TPointerFIFO;
	TPointerFIFO=object(TObject)
	private
		FCapacity,FHead,FTail,FUsed: Word;
		FData: Pointer;
		procedure SetCapacity(New: Word);
	public
		function Count: Word;
		procedure Put(P: Pointer);
		function Get: Pointer;
		destructor Done; virtual;
	end;
{Message functions}
	PMessageObject=^TMessageObject;
	PMessage=^TMessage;
	TMessage=object(TObject)
		Sender: PMessageObject;
		Message: Word;
	end;
	TMessageObject=object(TObject)
	private
		FMessages: TPointerFIFO;
		procedure CallDynamic(Dynamic: Integer; Param: Pointer);
	public
		constructor Init;
		destructor Done; virtual;
		procedure PostMessage(AMessage: PMessage);
		procedure Perform(AMessage: PMessage);
		procedure ProcessMessages;
	end;
{Thread functions}
	PThread=^TThread;
	TThread=object(TMessageObject)
	private
		FStack: Pointer;
		FStackSize,FSS: Word;
		FSP: Longint;
		FQueue: PMessage;
		FExcept: Pointer;
{		FPriority,FIdle: Word;}
		FMagic: LongInt;
		FNext,FPrev: PThread;
		procedure UnHook;
		procedure UnHookEx;
		procedure Hook(AThread: PThread);
		procedure Destroy;
	public
		Name: string[31];
		constructor Init(const AName: string; AStackSize: Word; User: Pointer);
		procedure Execute(User: Pointer); virtual;
		procedure Stop;
		procedure Start;
		function Running: Boolean;
{		procedure SetPriority(Priority: Integer);}
		destructor Done; virtual;
	end;
{Pipe functions}
	PPipeData=^TPipeData;
	TPipeData=record
		Size,Left: Word;
		Owner: PThread;
		Next: PPipeData;
		Data: record end;
	end;
	PPipe=^TPipe;
	TPipe=object(TObject)
	private
		FFirst,FLast: PPipeData;
		procedure SetData(const AData; Count: Integer);
		procedure GetData(var AData; var Count: Integer);
		procedure Assign(var F: Text);
	public
		I,O: Text;
		constructor Init;
		function Empty: Boolean;
		destructor Done; virtual;
	end;
{Locking objects}
	PMutex=^TMutex;
	TMutex=object(TObject)
	private
		FRef: Word;
		FWaiting: TPointerFIFO;
	public
		constructor Init;
		destructor Done; virtual;
		function Locked: Boolean;
		procedure Lock;
		procedure UnLock;
	end;

const
	EAbort=0;
	SwitchOk: Integer=1;

{MS timer}
var
	Timer: Longint;
	Threads: Integer;
	Main: TThread;
	Exception: Integer absolute ExitCode;

{CRT replacment routines}
procedure Sound(Hz: Word);
procedure NoSound;
procedure Delay(Ms: Longint);
function ReadKey: Char;
function KeyPressed: Boolean;
procedure ClearKeys;

{TRY..EXCEPT handler}
function Try: Boolean;
procedure EndTry;
procedure Raise;
procedure Abort;

{Thread support}
procedure Switch;
procedure SwitchTo(AThread: PThread);
procedure Forbid;
procedure Permit;
function self: PThread;

implementation

uses DOS;

type
	PExcept=^TExcept;
	TExcept=packed record
		SSP,SBP: Word;
		Addr: Pointer;
		Prev: PExcept;
	end;
	PPointerArray=^TPointerArray;
	TPointerArray=array[0..16382] of Pointer;

const
	GrowSize=5;
	TimerFreq=$1234DC;
	ThreadMagic=TimerFreq;
	Divider=8192; {145 Hz}
	Frequency=TimerFreq div Divider;
	CallOldInt8Base=65536 div Divider; {Every 8 calls one BIOS call}
	CallOldInt8: Word=1;

var
	TimerComma: Word;
	Current: PThread;
	OldInt8,OldInt9,LockInt16,OldExitMain,OldExitExcept: Pointer;
{$IFNDEF DPMI}
	InDOS: ^Byte;
{$ENDIF}
	ScanCode: Byte;

procedure TMessageObject.CallDynamic(Dynamic: Integer; Param: Pointer); assembler;
const
	vmtDMTPtr=4;
	dmtParent=0;
	dmtCacheIndex=2;
	dmtCacheEntry=4;
	dmtEntryCount=6;
	dmtEntryTable=8;
asm
	mov ax,Dynamic
	les di,[self]
	mov di,es:[di]
	mov bx,[di+vmtDMTPtr]
	cmp ax,[bx+dmtCacheIndex]
	jne @notcached
	mov di,[bx+dmtCacheEntry]
	jmp @docall
@notcached:
	mov si,ds
	mov es,si
	mov si,bx
	cld
@lookup:
	mov cx,[bx+dmtEntryCount]
	mov dx,cx
	lea di,[bx+dmtEntryTable]
	repne scasw
	jz @stop
	mov bx,es:[bx+dmtParent]
	or bx,bx
	jnz @lookup
	mov ExitCode,210
	jmp Raise
@stop:
	dec dx
	shl dx,1
	sub dx,cx
	shl dx,1
	add di,dx
	mov [si+dmtCacheIndex],ax
	mov [si+dmtCacheEntry],di
@docall:
db $66; push word ptr Param
db $66; push word ptr Self
	call dword ptr [di] {call with the own stack parameters}
end;

constructor TMessageObject.Init;
begin
	inherited Init;
	FMessages.Init;
end;

destructor TMessageObject.Done;
begin
	FMessages.Done;
	inherited Done;
end;

procedure TMessageObject.PostMessage(AMessage: PMessage);
begin
	FMessages.Put(AMessage);
end;

procedure TMessageObject.Perform(AMessage: PMessage);
begin
	if Try then begin
		CallDynamic(AMessage^.Message,AMessage);
		AMessage^.Free;
		EndTry;
	end else begin
		if Exception<>210 then
			Raise;
		AMessage^.Free;
	end;
end;

procedure TMessageObject.ProcessMessages;
var
	M: PMessage;
begin
	repeat
		M:=FMessages.Get;
		if not Assigned(M) then
			Break;
		Perform(M);
	until False;
	Switch;
end;

{*** TTHREAD ***}
procedure SelectThread; near;
var
	Act,Buf: PThread;
begin
	Buf:=Current;
	Buf^.UnHook;
	Act:=Current;
	if Assigned(Act) then begin
		while Act^.FNext<>nil do
			Act:=Act^.FNext;
	end;
	Buf^.Hook(Act);
end;

procedure SwitchThreadEnd; forward;

procedure SwitchThread; assembler;
asm
	cli
db $66;	push bx
db $66;	push cx
db $66;	push dx
db $66;	push si
db $66;	push di
db $66;	push bp
	push es
db $f,$a0 {push	fs}
db $f,$a8 {push	gs}
	push InOutRes
	push DosError
db $66; push word ptr Current
	call SelectThread
	mov bx,sp
	les	di,ss:[bx]
	add sp,4 {correction of Current-Push}
	mov	es:[di+TThread.FSS],ss
db $66;	mov	word ptr es:[di+TThread.FSP],sp
	les	di,Current
	mov	ss,es:[di+TThread.FSS]
db $66;	mov	sp,word ptr es:[di+TThread.FSP]
	jmp SwitchThreadEnd
end;

procedure SwitchThreadEnd; assembler;
asm
	pop	DosError
	pop	InOutRes
db $f,$a9{pop	gs}
db $f,$a1{pop	fs}
	pop	es
db $66;	pop	bp
db $66;	pop	di
db $66;	pop	si
db $66;	pop	dx
db $66;	pop	cx
db $66;	pop	bx
db $66;	pop	ax
	pop	ds
	sti
	iret
end;

procedure TThread.UnHookEx;
begin
	UnHook;
	if not Assigned(Current) then
		RunError(257); {Cannot unhook ALL threads!}
end;

procedure TThread.UnHook;
begin
	Forbid;
	if Assigned(FNext) then
		FNext^.FPrev:=FPrev;
	if Assigned(FPrev) then begin
		FPrev^.FNext:=FNext;
		FPrev:=nil;
	end else
		Current:=FNext;
	FNext:=nil;
	Permit;
end;

procedure TThread.Hook(AThread: PThread);
begin
	Forbid;
	if Assigned(AThread) then begin
		FNext:=AThread^.FNext;
		FPrev:=AThread;
		AThread^.FNext:=@self;
		if Assigned(FNext) then
			FNext^.FPrev:=@self;
	end else begin
		FNext:=Current;
		Current:=@self;
		if Assigned(FNext) then
			FNext^.FPrev:=@self;
	end;
	Permit;
end;

procedure TThread.Destroy;
	procedure FreeMemory;
	var
		Act: PExcept;
	begin
		Freemem(FStack,FStackSize);
		repeat
			Act:=FExcept;
			if not Assigned(Act) then
				Break;
			FExcept:=Act^.Prev;
			Dispose(Act);
		until False;
	end;
begin
	Forbid;
	if not Assigned(FStack) then
		Halt(0); {This was the main thread, no MAGIC check!}
	if FMagic<>ThreadMagic then
		RunError(257); {This ain't a thread!!!}
	FMagic:=0;
	Dec(Threads);
	if @self<>Current then begin
		if Running then
			UnHookEx;
		FreeMemory;
		if Seg(self)<>DSeg then
			Dispose(@self);
	end else begin
		UnHookEx;
		asm
			cli
		db $66; mov ax,word ptr self
			les di,Current
			mov	ss,es:[di+TThread.FSS]
		db $66;	mov	sp,word ptr es:[di+TThread.FSP]
		db $66; push ax
			mov bp,sp{Self}
			sub bp,6 {Ret-Address, BP-Save}
			sti
		end;
		FreeMemory;
		if Seg(self)<>DSeg then
			Dispose(@self);
		asm
			add sp,4 {Self}
			call Permit
			jmp SwitchThreadEnd
		end;
	end;
end;

constructor TThread.Init(const AName: string; AStackSize: Word; User: Pointer);
{$IFDEF DPMI}
const
	pStackTop=$0A;
{	pStackBot=$0C;}
	pStackMin=$0E;
{$ENDIF}
begin
	inherited Init;
	Forbid;
	Name:=AName;
	FMagic:=ThreadMagic;
	FStackSize:=AStackSize and $FFFC; {4-Byte boundary}
	GetMem(FStack,FStackSize);
	FSS:=Seg(FStack^);
	FSP:=Ofs(FStack^)+FStackSize;
	asm
		cld
		pushf
		cli
		pop cx
		mov dx,ss
	db $66; mov si,sp
		les di,self
	db $66; mov bx,word ptr User
		mov ss,es:[di].FSS
	db $66; mov sp,word ptr es:[di].FSP
	{$IFDEF DPMI}
		mov ax,es:[di].FStackSize
		mov ss:[pStackMin],ax
		sub ax,ax
		mov ss:[pStackTop],ax
	{$ENDIF}
	db $66; sub ax,ax
		push es       {=Object segment}
		push di       {=Object Offset}
	db $66; push ax {=Fake return address}
	db $66; push bx {=USER-Parameter}
		push es       {=Object segment}
		push di       {=Object offset}
		push seg TThread.Free {=Free codeseg}
		push offset TThread.Free {=Free offset}
		push cx       {=Flags}
		mov bx,es:[di]
	db $66; push word ptr ds:[bx+4+vmtHeaderSize] {=Execute address from VMT}
		push ds				{= DS=DS}
	db $66;	push ax	{=EAX= 0}
	db $66;	push ax	{=EBX= 0}
	db $66;	push ax	{=ECX= 0}
	db $66;	push ax	{=EDX= 0}
	db $66;	push ax	{=ESI= 0}
	db $66;	push ax	{=EDI= 0}
	db $66;	push ax	{=EBP= 0}
		push ax 			{= ES= 0}
		push ax 			{= FS= 0}
		push ax 			{= GS= 0}
		push ax 			{=INOUTRES=0}
		push ax 			{=DOSERROR=0}
	db $66; mov	word ptr es:[di].FSP,sp
	db $66; mov sp,si
		mov ss,dx
		sti
	end;
	Inc(Threads);
	Hook(Current);
	Permit;
end;

procedure TThread.Execute;
begin
	Abstract;
end;

function TThread.Running: Boolean;
begin
	Running:=Assigned(FPrev) or Assigned(FNext) or (@self=Current);
end;

procedure TThread.Stop;
var
	SelfStop: Boolean;
begin
	if Assigned(FPrev) or Assigned(FNext) then begin
		SelfStop:=@self=Current;
		UnHookEx;
		if SelfStop then
			asm
				pushf
				cli
				push cs
				push offset @restart
				push ds
			db $66;	push ax
			db $66;	push bx
			db $66;	push cx
			db $66;	push dx
			db $66;	push si
			db $66;	push di
			db $66;	push bp
				push es
			db $f,$a0 {push	fs}
			db $f,$a8 {push	gs}
				push InOutRes
				push DosError
				les	di,Self
				mov	es:[di+TThread.FSS],ss
			db $66;	mov	word ptr es:[di+TThread.FSP],sp
				les	di,Current
				mov	ss,es:[di+TThread.FSS]
			db $66;	mov	sp,word ptr es:[di+TThread.FSP]
				jmp SwitchThreadEnd
			@restart:
			end;
	end;
end;

procedure TThread.Start;
begin
	if not (Assigned(FPrev) or Assigned(FNext) or (Current=@self)) then
		Hook(Current);
end;

destructor TThread.Done;
begin
	Destroy;
end;

{*** TPIPE ***}
function PipeTextNop(var F: TextRec): Integer; far; assembler;
asm
	sub ax,ax
end;

function PipeTextIn(var F: TextRec): Integer; far; assembler;
asm
	inc SwitchOk
	les di,F
db $66; push word ptr es:[di+TextRec.BufPtr]
	push es
	mov ax,di
	add ax,TextRec.BufEnd
	push ax
	mov es:[di+TextRec.BufPos],0
db $66; push word ptr es:[di+TextRec.UserData]
	call far ptr TPipe.GetData
	sub ax,ax
	dec SwitchOk
end;

function PipeTextOut(var F: TextRec): Integer; far; assembler;
asm
	inc SwitchOk
	les di,F
	sub ax,ax
	xchg ax,es:[di+TextRec.BufPos]
	cmp ax,0
	je @end
db $66; push word ptr es:[di+TextRec.BufPtr]
	push ax
db $66; push word ptr es:[di+TextRec.UserData]
	call far ptr TPipe.SetData
	sub ax,ax
@end:
	dec SwitchOk
end;

function PipeTextOpen(var F: TextRec): Integer; far;
begin
	with F do begin
		case Mode of
		fmInput: begin
				InOutFunc:=@PipeTextIn;
				FlushFunc:=@PipeTextNop;
			end;
		fmOutput: begin
				InOutFunc:=@PipeTextOut;
				FlushFunc:=@PipeTextOut;
			end;
		else
			PipeTextOpen:=12; {Invalid access code}
			Exit;
		end;
		CloseFunc:=@PipeTextNop;
		PipeTextOpen:=0;
	end;
end;

procedure TPipe.SetData(const AData; Count: Integer);
var
	ASize: Integer;
begin
	ASize:=SizeOf(TPipeData)+Count;
	if Assigned(FLast) then begin
		GetMem(FLast^.Next,ASize);
		FLast:=FLast^.Next;
	end else begin
		GetMem(FFirst,ASize);
		FLast:=FFirst;
	end;
	with FLast^ do begin
		Size:=ASize;
		Left:=Count;
		Owner:=Current;
		Next:=nil;
		Move(AData,Data,Count);
	end;
end;

procedure TPipe.GetData(var AData; var Count: Integer);
var
	Act,Prev: PPipeData;
begin
	Prev:=nil;
	Act:=FFirst;
	while Assigned(Act) and (Act^.Owner=Current) do begin
		Prev:=Act;
		Act:=Act^.Next;
	end;
	if Assigned(Act) then begin
		with Act^ do begin
			Count:=Left;
			Move(Data,AData,Count);
		end;
		if Assigned(Prev) then
			Prev^.Next:=Act^.Next
		else
			FFirst:=Act^.Next;
		if Act=FLast then
			FLast:=Prev;
		FreeMem(Act,Act^.Size);
	end else
		Count:=0;
end;

function TPipe.Empty: Boolean;
var
	Act: PPipeData;
begin
	Act:=FFirst;
	while Assigned(Act) and (Act^.Owner=Current) do
		Act:=Act^.Next;
	Empty:=not Assigned(Act);
end;

procedure TPipe.Assign(var F: Text);
begin
	with TextRec(F) do begin
		Mode:=fmClosed;
		BufSize:=SizeOf(Buffer);
		BufPtr:=@Buffer;
		OpenFunc:=@PipeTextOpen;
		Name[0]:=#0;
		asm
			les di,F
		db $66; mov ax,word ptr self
		db $66; mov word ptr es:[di+TextRec.UserData],ax
		end;
	end;
end;

constructor TPipe.Init;
begin
	inherited Init;
	Assign(I);
	Reset(I);
	Assign(O);
	Rewrite(O);
end;

destructor TPipe.Done;
begin
	Close(I);
	Close(O);
	inherited Done;
end;

procedure TPointerFIFO.SetCapacity(New: Word);
var
	Buf: Pointer;
	I: Integer;
begin
	if New<FUsed then
		RunError(204);
	if New=0 then
		FreeMem(FData,FUsed*SizeOf(Pointer))
	else begin
		GetMem(Buf,New*SizeOf(Pointer));
		if Assigned(FData) then begin
			if (FUsed>0) then
				if FTail>FHead then begin
					I:=FCapacity-FTail;
					Move(PPointerArray(FData)^[FTail],Buf^,Pred(I)*SizeOf(Pointer));
					Move(FData^,PPointerArray(Buf)^[I],FHead*SizeOf(Pointer));
				end else begin
					Move(PPointerArray(FData)^[FTail],Buf^,FUsed*SizeOf(Pointer));
				end;
			FreeMem(FData,FUsed*SizeOf(Pointer));
		end;
	end;
	FCapacity:=New;
	FData:=Buf;
	FHead:=FUsed;
	FTail:=0;
end;

function TPointerFIFO.Count: Word; assembler;
asm
	les di,self
	mov ax,es:[di].FUsed
end;

procedure TPointerFIFO.Put(P: Pointer);
begin
	if Assigned(P) then begin
		Forbid;
		if FCapacity=FUsed then
			SetCapacity(FCapacity+GrowSize);
		PPointerArray(FData)^[FHead]:=P;
		FHead:=(FHead+1) mod FCapacity;
		Inc(FUsed);
		Permit;
	end;
end;

function TPointerFIFO.Get: Pointer;
begin
	if FUsed=0 then
		Get:=nil
	else begin
		Forbid;
		if FCapacity-FUsed>10*GrowSize then
			SetCapacity(FUsed+GrowSize);
		Dec(FUsed);
		Get:=PPointerArray(FData)^[FTail];
		FTail:=(FTail+1) mod FCapacity;
		Permit;
	end;
end;

destructor TPointerFIFO.Done;
begin
	FUsed:=0;
	SetCapacity(0);
	inherited Done;
end;

constructor TMutex.Init;
begin
	inherited Init;
	FWaiting.Init;
end;

destructor TMutex.Done;
begin
	FWaiting.Done;
	inherited Done;
end;

function TMutex.Locked: Boolean;
begin
	Locked:=FRef>0;
end;

procedure TMutex.Lock;
begin
	Inc(FRef);
	if FRef>1 then begin
		FWaiting.Put(Current);
		Current^.Stop;
	end;
end;

procedure TMutex.UnLock;
var
	Buf: PThread;
begin
	Buf:=FWaiting.Get;
	if Assigned(Buf) then
		Buf^.Start;
	Dec(FRef);
end;

{*** CRT ***}
procedure Sound(Hz: Word); assembler;
asm
	mov bx,Hz
	mov ax,$34DD
	mov dx,$12
	cmp dx,bx
	jnc	@end
	div bx
	mov bx,ax
	in al,$61
	test al,3
	jnz @setfreq
	or al,3
	out $61,al
	mov al,$B6
	out $43,al
@setfreq:
	mov al,bl
	out $42,al
	mov al,bh
	out $42,al
@end:
{$IFDEF DPMI}
	call Switch
{$ENDIF}
end;

procedure NoSound; assembler;
asm
	in al,$61
	and al,$fc
	out $61,al
{$IFDEF DPMI}
	call Switch
{$ENDIF}
end;

procedure Delay(Ms: Longint); assembler;
asm
db $66; mov ax,word ptr Ms
db $66; add ax,word ptr Timer
	jmp @start
@loop:
	call Switch
@start:
db $66; cmp word ptr Timer,ax
	jb @loop
end;

function ReadKey: Char; assembler;
asm
{$IFDEF DPMI}
	call Switch
{$ENDIF}
	mov al,ScanCode
	mov ScanCode,0
	cmp al,0
	jne	@end
@loop:
	call KeyPressed
	cmp al,False
	jne @cont
	call Switch
	jmp @loop
@cont:
	mov ah,0
	int $16
	cmp al,0
	jne @end
	mov ScanCode,ah
@end:
end;

function KeyPressed: Boolean; assembler;
asm
{$IFDEF DPMI}
	call Switch
{$ENDIF}
	mov al,ScanCode
	cmp al,0
	jne @true
	mov ah,1
	int $16
	mov al,0
	je @end
@true:
	mov al,True
@end:
end;

procedure ClearKeys;
begin
	while KeyPressed do
		ReadKey;
end;

{*** TRY...EXCEPT ***}
procedure CreateTry;
begin
	New(PExcept(Current^.FExcept));
end;

procedure DestroyTry;
begin
	Dispose(PExcept(Current^.FExcept));
end;

procedure NewExitExcept; far; assembler;
asm
db $66; cmp word ptr ErrorAddr,0
	je @end {Normal exit}
	les di,Current
db $66; cmp word ptr es:[di+TThread.FExcept],0
	je @end {No handler installed}
	les di,es:[di+TThread.FExcept]
{HERE THE STACK TEST COULD BE INSERTED}
	mov sp,es:[di+TExcept.SSP]
	mov bp,es:[di+TExcept.SBP]
db $66; push word ptr es:[di+TExcept.Addr]
db $66; push word ptr es:[di+TExcept.Prev]
	call DestroyTry
	les di,Current
db $66; pop word ptr es:[di+TThread.FExcept]
	mov word ptr ExitProc+2,seg NewExitExcept
	mov word ptr ExitProc,offset NewExitExcept
	mov al,False
	retf
@end:
db $66; mov ax,word ptr OldExitExcept
db $66; mov word ptr ExitProc,ax
end;

function Try: Boolean; assembler;
asm
db $66; cmp word ptr OldExitExcept,0
	jnz @installed
db $66; mov ax,word ptr ExitProc
db $66; mov word ptr OldExitExcept,ax
	mov word ptr ExitProc+2,seg NewExitExcept
	mov word ptr ExitProc,offset NewExitExcept
@installed:
	les di,Current
db $66; push word ptr es:[di+TThread.FExcept]
	call CreateTry
	les di,Current
	les di,es:[di+TThread.FExcept]
db $66; pop word ptr es:[di+TExcept.Prev]
	mov bx,sp
db $66; mov ax,word ptr ss:[bx]
	add bx,4
	mov es:[di+TExcept.SSP],bx
	mov es:[di+TExcept.SBP],bp
db $66; mov word ptr es:[di+TExcept.Addr],ax
	mov al,True
end;

procedure EndTry; assembler;
asm
	les di,Current
	les di,es:[di+TThread.FExcept]
db $66; push word ptr es:[di+TExcept.Prev]
	call DestroyTry
	les di,Current
db $66; pop word ptr es:[di+TThread.FExcept]
end;

procedure Raise;
begin
	RunError(ExitCode);
end;

procedure Abort;
begin
	RunError(0);
end;

{*** THREADING ***}
procedure Switch; assembler;
asm
	pushf
	push cs
	push offset @end
	push ds
db $66;	push	ax
	mov ax,ds
	mov ds,ax
	test SwitchOk,-1
	jnz @pop
	jmp	SwitchThread
@pop:
db $66;	pop	ax
	pop ds
	add sp,4 {kill return address}
	popf
@end:
end;

procedure SwitchTo(AThread: PThread);
begin
	Forbid;
	AThread^.UnHook;
	AThread^.Hook(Current);
	Permit;
	Switch;
end;

procedure Forbid; assembler;
asm
	inc SwitchOk
	cmp SwitchOk,-1
	jne @end
	dec SwitchOk
@end:
end;

procedure Permit; assembler;
asm
	cmp SwitchOk,0
	je @end
	dec SwitchOk
@end:
end;

function self: PThread; assembler;
asm
	mov ax,word ptr Current
	mov dx,word ptr Current+2
end;

{*** LOW LEVEL ***}
{New timer interrupt}
procedure NewInt8; far; assembler;
const
	TimerAddFull=Round(66536000/(TimerFreq/Divider));
	TimerAddInt=TimerAddFull div 65536;
	TimerAddComma=TimerAddFull mod 65536;
asm
	push	ds
db $66;	push	ax
	mov	ax,seg @data
	mov	ds,ax
	add word ptr TimerComma,TimerAddComma
db $66;	adc	word ptr Timer,TimerAddInt
	dec	CallOldInt8
	jnz	@nocall
	mov CallOldInt8,CallOldInt8Base
	pushf
	call	[OldInt8]
	jmp	@endcall
@nocall:
	mov	al,$20
	out	$20,al
@endcall:
{$IFNDEF DPMI}
	test SwitchOk,-1
	jnz	@end
	mov ax,bx
	lds	bx,InDOS
	cmp	byte ptr ds:[bx],0
	mov bx,ax
	jne	@end
	mov	ax,seg @data
	mov	ds,ax
	jmp	SwitchThread
{$ENDIF}
@end:
db $66;	pop ax
	pop	ds
	iret
end;

{New KDB-Interrupt: Lock during interrupt}
procedure NewInt9; far; assembler;
asm
	push ds
	push seg @data
	pop	ds
	inc	SwitchOk
	pushf
	call OldInt9
	dec	SwitchOk
	pop	ds
	iret
end;

{Universal Lock-And-Run for Interrupts. Flags returned by interrupt are used}
{*** SPECIAL COPY-MODIFY-EXECUTE HANDLER. DO NOT CHANGE ANYTHING! ***}
const
	NewIntLockSize=$32;
	NewIntCallAddr=$11;

procedure NewIntLock; far; assembler; {DO NOT MODIFY!}
asm
	pushf
	pushf
	push ds
	push bx
	mov bx,seg @data
	mov ds,bx
	inc	SwitchOk
	pop bx
	pop ds
	popf
	db $9A; dd $FFFFFFFF
	push ds
	push ax
	push bx
	pushf
	mov bx,seg @data
	mov ds,bx
	dec	SwitchOk
	pop ax
	mov bx,sp
	mov ss:[bx+10],ax {Transfer flags}
	pop bx
	pop ax
	pop ds
	call Switch
	iret
@intnr:
end;

function InitIntLock(AInt: Byte): Pointer;
var
	Buf,OldInt: Pointer;
begin
	Forbid;
	GetIntVec(AInt,OldInt);
	GetMem(Buf,NewIntLockSize+1);
	Move(@NewIntLock^,Buf^,NewIntLockSize);
	asm
		les di,Buf
	db $66; mov ax,word ptr OldInt
	db $66; mov es:[di+NewIntCallAddr],ax
		mov al,AInt
		mov es:[di+NewIntLockSize],al
	end;
	SetIntVec(AInt,Buf);
	InitIntLock:=Buf;
	Permit;
end;

procedure DoneIntLock(ALock: Pointer);
var
	OldInt: Pointer;
	AInt: Byte;
begin
	if not Assigned(ALock) then
		Exit;
	Forbid;
	asm
		les di,ALock
	db $66; mov ax,es:[di+NewIntCallAddr]
	db $66; mov word ptr OldInt,ax
		mov al,es:[di+NewIntLockSize]
		mov AInt,al
	end;
	SetIntVec(AInt,OldInt);
	FreeMem(ALock,NewIntLockSize+1);
	Permit;
end;
{*** SPECIAL HANDLER CODE ENDS HERE. ***}

{*** START/END CODE ***}
{Set timer frequency}
procedure SetTimer(Divider: Longint); assembler;
asm
	mov al,$36
	out $43,al
	mov ax,word ptr Divider
	out $40,al
	mov al,ah
	out $40,al
	mov al,$20
	out $20,al
end;

procedure NewExitMain; far;
begin
	SwitchOk:=1;
	ExitProc:=OldExitMain;
	SetTimer(65536);
	SetIntVec($8,OldInt8);
	{$IFNDEF DPMI}
	SetIntVec($9,OldInt9);
	DoneIntLock(LockInt16);
	{$ENDIF}
end;

begin
	{$IFNDEF DPMI}
	asm
		mov	ah,$34
		int	$21
		mov	word ptr InDOS,bx
		mov	word ptr InDOS+2,es
	end;
	{$ENDIF}
	FillChar(Main,SizeOf(Main),0);
	Current:=@Main;
	OldExitMain:=ExitProc;
	ExitProc:=@NewExitMain;
	GetIntVec($8,OldInt8);
	SetIntVec($8,@NewInt8);
	{$IFNDEF DPMI}
	GetIntVec($9,OldInt9);
	SetIntVec($9,@NewInt9);
	LockInt16:=InitIntLock($16);
	{$ENDIF}
	SetTimer(Divider);
	Threads:=1;
	SwitchOk:=0;
	ScanCode:=0;
	Switch; {Inogurate MAIN}
end.
